//----------------------------------------------------------------------------
//	HEADER:	Map.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Header file of Map demo application
//----------------------------------------------------------------------------

#ifndef _MAP_H
#define _MAP_H

//////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#define STRICT
#define NOMINMAX
#define NONAMELESSUNION
#define _WIN32_WINNT 0x0400
#include <tchar.h>
#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <math.h>

// Ignore messages about inlining
#pragma warning(disable: 4711)

// Release-only stuff
// WIBBLE #define MAPDLL_HASH
#ifdef _DEBUG
#define local
#else
#define local static
#endif

// Operating system calls
#define MAPDLL_STD WINAPI
#define MAPDLL_ASC WINAPI
#define MAPDLL_VAR WINAPIV

#ifdef MAPDLL_HASH

// DLL entry point hashing
extern void DLLLoad(void);
#define MAPDLL(std, lib, ret, fnc, prm, hsh)	typedef ret (MAPDLL_##std* OS_##fnc##_Ptr)prm; OS_##fnc##_Ptr OS_##fnc;
#include "MapDLL.h"
#undef MAPDLL
#define OS(func)				OS_##func

#else

// No DLL entry point hashing
#pragma comment(lib, "COMCTL32")
#define OS(func)				func

#endif

//////////////////////////////////////////////////////////////////////////////

// Debug or Release
#ifdef _DEBUG
BOOL DebugAssert(LPCTSTR file, int line, LPCTSTR reason);
#define TRACE(why)				DebugAssert(_T(__FILE__), __LINE__, why)
#define BREAKPOINT(why)			if (TRACE(why)) __asm { int 3 } else 0
#define ASSERT(expr)			if (!(expr)) BREAKPOINT(_T("Assertion failure: ") _T(#expr)); else __assume(expr)
#define VERIFY(expr)			if (!(expr)) BREAKPOINT(_T("Verify failure: ") _T(#expr)); else 0
#define UNREACHABLE()			BREAKPOINT(_T("Unreachable code reached"))
#else
#define ASSERT(expr)			__assume(expr)
#define VERIFY(expr)			expr
#define UNREACHABLE()			__assume(0)
#endif
#define USE(var)				while (&(var) != &(var))

// The instruction list
#define MAX_INSTRUCTIONS		2048
extern int raw_instruction[MAX_INSTRUCTIONS];
extern int raw_instructions;

// The area list
#define MAX_AREAS				512
typedef struct _MAPAREA
{
	int code;
	int point_offset;
}
MAPAREA;
extern MAPAREA raw_area[MAX_AREAS];
extern int raw_areas;

// The point list
#define MAX_POINTS				65536
typedef struct _MAPPOINT
{
	float x;
	float y;
}
MAPPOINT;
extern MAPPOINT raw_point[MAX_POINTS];
extern int raw_points;

// The colourising callback
typedef COLORREF (*MAPFUNC)(int area, int rx, int ry);
#define RGB_NULL 0xFFFFFFFF
#define RGB_SEA PALETTERGB(0x00, 0x00, 0xFF)
#define RGB_BLACK PALETTERGB(0x00, 0x00, 0x00)

//////////////////////////////////////////////////////////////////////////////

// Constants (we hope)
#define M_PI					3.1415926535897932384626433832795
#define M_SIN_45				0.7071067811865475244008443621049
#define M_SIN_60				0.8660254037844386467637231707529

// Useful little macro
#define dimof(x)				(sizeof(x) / sizeof(*(x)))

//////////////////////////////////////////////////////////////////////////////

// MapArea.c
extern LPCTSTR region_names[];
extern void Decompress(void);
extern int RegionToCountry(int region);

// MapProj.c
extern void ProjectionDecompress(void);

// MapTool.c
extern HWND TooltipCreate(HWND hparent);
extern void TooltipTrack(LPPOINT pt, int region);

// MapRep.c
extern HWND ReportCreate(HWND hparent, HICON hicon);
extern void ReportSelect(int country);
extern LPTSTR ReportTooltip(int region);
extern BOOL ReportSaveAs(LPCTSTR path);

// MapRep.c
extern void MapRecolour(void);
extern void MapRepaint(void);
extern COLORREF ColourDefault(int area, int rx, int ry);
extern COLORREF ColourTimeZone(int area, int rx, int ry);
extern COLORREF ColourGrey(int area, int rx, int ry);

// MapFlag.c
extern void FlagDecompress(void);
extern double FlagAspectRatio(int flag);
extern void FlagDraw(HDC hdc, int width, int height, double aspect, int flag);

// Map.c
extern void FlagSelect(int country);
extern void RegionZoom(int country);

//////////////////////////////////////////////////////////////////////////////

#endif

//////////////////////////////////////////////////////////////////////////////
