//----------------------------------------------------------------------------
//	SOURCE:	MapDLL.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Map demo cutdown import library
//			From original work by Marcus Winter at http://www.20to4.net
//----------------------------------------------------------------------------

#include "Map.h"

// Sanity check
#define MAPDLL_LINKAGE_KERNEL32	DECLSPEC_IMPORT
#define MAPDLL_LINKAGE_USER32	DECLSPEC_IMPORT
#define MAPDLL_LINKAGE_GDI32	DECLSPEC_IMPORT
#define MAPDLL_LINKAGE_COMCTL32	DECLSPEC_IMPORT
#define MAPDLL_LINKAGE_COMDLG32	DECLSPEC_IMPORT
#define MAPDLL(std, lib, ret, fnc, prm, hsh)	MAPDLL_LINKAGE_##lib ret MAPDLL_##std fnc prm;
#include "MapDLL.h"
#undef MAPDLL

#ifdef MAPDLL_HASH

local DWORD DLLHash(LPCTSTR func)
{
	// Calculate a hash for a string
	DWORD hash = *func;
	while (*++func != '\0')
	{
		hash = hash * 31 + *func;
	}
	return hash;
}

local FARPROC DLLLoadFunction(HINSTANCE inst, DWORD hash)
{
	// Scan the module's import table for a string hash match
	const char* libmem = (const char*)inst;
	const IMAGE_DOS_HEADER* doshdr = (const IMAGE_DOS_HEADER*)libmem;
	const IMAGE_NT_HEADERS* nthdr = (const IMAGE_NT_HEADERS*)(libmem + doshdr->e_lfanew);
	const IMAGE_EXPORT_DIRECTORY* expdir = (const IMAGE_EXPORT_DIRECTORY*)(libmem + nthdr->OptionalHeader.DataDirectory[0].VirtualAddress);
	const DWORD* names = (const DWORD*)(libmem + expdir->AddressOfNames);
	const char* name = libmem + *names;
	DWORD i = expdir->NumberOfNames;
	while (hash != DLLHash(name))
	{
		i--;
		ASSERT(i > 0);
		name = libmem + *++names;
	}
	return GetProcAddress(inst, name);
}

#ifdef _DEBUG

local FARPROC DLLLoadFunctionEx(HINSTANCE inst, const char* func, const char* output, DWORD oldhash)
{
	// Report the hash for a function in a copy-paste stylee
	FARPROC address = GetProcAddress(inst, func);
	DWORD newhash = DLLHash(func);
	char buffer[256];
	wsprintfA(buffer, output, newhash);
	OutputDebugStringA(buffer);
	Sleep(0);
	ASSERT(address != NULL);
	ASSERT(address == DLLLoadFunction(inst, newhash));
	ASSERT(oldhash == newhash);
	return address;
}

#endif

void DLLLoad(void)
{
	HINSTANCE KERNEL32 = LoadLibrary("KERNEL32");
	HINSTANCE USER32 = LoadLibrary("USER32");
	HINSTANCE GDI32 = LoadLibrary("GDI32");
	HINSTANCE COMCTL32 = LoadLibrary("COMCTL32");
	HINSTANCE COMDLG32 = LoadLibrary("COMDLG32");

// Unicode/ASCII suffixes
#define MAPDLL_SUFFIX_STD
#ifdef UNICODE
#define MAPDLL_SUFFIX_ASC	"W"
#else
#define MAPDLL_SUFFIX_ASC	"A"
#endif
#define MAPDLL_SUFFIX_VAR	MAPDLL_SUFFIX_ASC

// Load the required operating system functions
#ifdef _DEBUG
#define MAPDLL(std, lib, ret, fnc, prm, hsh)	OS_##fnc = (OS_##fnc##_Ptr)DLLLoadFunctionEx(lib, #fnc MAPDLL_SUFFIX_##std, "MAPDLL(" #std ", " #lib ", " #ret ", " #fnc ", " #prm ", 0x%08X)\r\n", hsh);
#else
#define MAPDLL(std, lib, ret, fnc, prm, hsh)	OS_##fnc = (OS_##fnc##_Ptr)DLLLoadFunction(lib, hsh);
#endif

#include "MapDLL.h"
}

#endif

