//----------------------------------------------------------------------------
//	SOURCE:	MapFlag.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Source file of Map2 demo flags
//----------------------------------------------------------------------------

#include "Map.h"
#include "MapProj.h"

#define COUNTRY(a,b,d)	Flag_##a,
enum Flag
{
	Flag_None = 0,
#include "MapTable.h"
	Flag_Count
};

// Flag colours
#define FLAG_COLOUR_LIST(X) \
    X(N,   0,0,0) \
    X(R,   F,0,0) \
    X(Rp,  C,0,0) \
    X(Rpp, 9,0,0) \
    X(V,   0,9,0) \
    X(Vp,  0,6,0) \
    X(Bm,  3,9,F) \
    X(B,   0,0,F) \
    X(Bpp, 0,0,9) \
    X(Y,   F,F,0) \
    X(Om,  F,9,0) \
    X(O,   F,6,0) \
    X(M,   6,3,0) \
    X(Gm,  C,C,C) \
    X(W,   F,F,F)
#define FLAG_COLOUR_LIST_ENUM(x,r,g,b) FlagColour_##x,
enum FlagColour
{
    FlagColour_None = -1,
    FLAG_COLOUR_LIST(FLAG_COLOUR_LIST_ENUM)
    FlagColour_Count
};

#define CMD_LIST(X) \
    X(Blockh) \
    X(Blockv) \
    X(Circle) \
    X(Colour) \
    X(Cross) \
    X(Curve) \
    X(Diagonal) \
    X(Flood) \
    X(Forr) \
    X(Fors) \
    X(Fort) \
    X(Star) \
    X(Star5) \
    X(Stripesh) \
    X(Stripesv)
#define CMDEXT_LIST(X) \
    X(Jack) \
    X(Triangle) \
    X(Rectangle) \
    X(Viewbox)
#define CMD_LIST_ENUM(x) Cmd_##x,
typedef enum Cmd
{
    Cmd_Ext = 0,
    CMD_LIST(CMD_LIST_ENUM)
    Cmd_Count
} Cmd;
#define CMD_MAKE(a,b)	(((BYTE)(a) << 4) + (BYTE)(b))
#define CMD_GET_CMD(ab)	((Cmd)((ab) >> 4))
#define CMD_GET_ARG(ab)	((BYTE)(ab) & 15)

#define CMDEXT_LIST_ENUM(x) CmdExt_##x,
enum CmdExt
{
    CmdExt_End = 0,
    CMDEXT_LIST(CMDEXT_LIST_ENUM)
    CmdExt_Count
};
#define Cmd_End CMD_MAKE(Cmd_Ext, CmdExt_End)
#define Cmd_EOF CMD_MAKE(Cmd_Ext, CmdExt_Count)

enum CmdCurve
{
	CmdCurve_End = -128,
	CmdCurve_Linear,
	CmdCurve_Smooth,
	CmdCurve_Cubic
};

#define FLAG_BEGIN(iso)
#define FLAG_END(iso)						Cmd_End,

#define COLOUR_EX(c)						FlagColour_##c
#define PAIR(a,b)							((a) + ((b) << 4))
#define TRIPLE_EX(a,b,c)					(((unsigned)(a+15) << 10) | ((unsigned)(b+15) << 5) | ((unsigned)(c+15) << 0))
#define TRIPLE(a,b,c)						(TRIPLE_EX(a,b,c) >> 8), (TRIPLE_EX(a,b,c) & 255) 

#define ASPECT_RATIO_MIN					(32.0 / 39.0) // Nepal NP
#define ASPECT_RATIO_MAX					(28.0 / 11.0) // Qatar QA

#define ASPECT_RATIO(x,y)					(BYTE)(((double)x / (double)y - ASPECT_RATIO_MIN) / (ASPECT_RATIO_MAX - ASPECT_RATIO_MIN) * 255),
#define BLOCK_H(c,y0,y1)					CMD_MAKE(Cmd_Blockh,COLOUR_EX(c)), y0, y1,
#define BLOCK_V(c,x0,x1)					CMD_MAKE(Cmd_Blockv,COLOUR_EX(c)), x0, x1,
#define CIRCLE(x,y,r)						CMD_MAKE(Cmd_Circle,r), x, y,
#define CIRCLE_L(x,y,r)						CMD_MAKE(Cmd_Circle,0), x, y, r,
#define COLOUR(c)							CMD_MAKE(Cmd_Colour,COLOUR_EX(c)),
#define CROSS(c,w,x)						CMD_MAKE(Cmd_Cross,COLOUR_EX(c)), w, x,
#define CURVE_BEGIN(x,y)					CMD_MAKE(Cmd_Curve,0), x, y,
#define CURVE_BEGIN_MIRROR(x,y)				CMD_MAKE(Cmd_Curve,1), x, y,
#define CURVE_LINEAR(x,y)					x, y,
#define CURVE_SMOOTH(x2,y2,x,y)				x2, y2, x, y,
#define CURVE_CUBIC(x1,y1,x2,y2,x,y)		x1, TRIPLE(y1,x2,y2), x, y,
#define CURVE_LINEAR_SET(x,y)				CmdCurve_Linear, CURVE_LINEAR(x,y)
#define CURVE_SMOOTH_SET(x2,y2,x,y)			CmdCurve_Smooth, CURVE_SMOOTH(x2,y2,x,y)
#define CURVE_CUBIC_SET(x1,y1,x2,y2,x,y)	CmdCurve_Cubic, CURVE_CUBIC(x1,y1,x2,y2,x,y)
#define CURVE_END()							CmdCurve_End,
#define DIAGONAL(lr,x)						CMD_MAKE(Cmd_Diagonal,lr), x,
#define FLOOD(c)							CMD_MAKE(Cmd_Flood,COLOUR_EX(c)),
#define FOR_ROTATE(n)						CMD_MAKE(Cmd_Forr,n),
#define FOR_ROTATE_L(n)						CMD_MAKE(Cmd_Forr,0), n,
#define FOR_SCALE(sx, sy)					CMD_MAKE(Cmd_Fors,sx+1), (signed char)(sy * 60.0),
#define FOR_TRANSLATE(n, dx,dy)				CMD_MAKE(Cmd_Fort,n), dx, dy,
#define JACK()								CMD_MAKE(Cmd_Ext,CmdExt_Jack),
#define RECTANGLE(x0,y0,x1,y1)				CMD_MAKE(Cmd_Ext,CmdExt_Rectangle), x0, y0, x1, y1,
#define STAR(n,r1,r2,x,y,rot)				CMD_MAKE(Cmd_Star,n-4), r1, r2, rot, x, y,
#define STAR_5S(x,y,r)						CMD_MAKE(Cmd_Star5,r-1), x, y,
#define STAR_5L(x,y,r)						CMD_MAKE(Cmd_Star5,0), r, x, y,
#define STRIPES_H2(n,c0,c1)					CMD_MAKE(Cmd_Stripesh,2), PAIR(n,COLOUR_EX(c0)), COLOUR_EX(c1),
#define STRIPES_H3(n,c0,c1,c2)				CMD_MAKE(Cmd_Stripesh,3), PAIR(n,COLOUR_EX(c0)), PAIR(COLOUR_EX(c1),COLOUR_EX(c2)),
#define STRIPES_H4(n,c0,c1,c2,c3)			CMD_MAKE(Cmd_Stripesh,4), PAIR(n,COLOUR_EX(c0)), PAIR(COLOUR_EX(c1),COLOUR_EX(c2)), COLOUR_EX(c3),
#define STRIPES_V2(n,c0,c1)					CMD_MAKE(Cmd_Stripesv,2), PAIR(n,COLOUR_EX(c0)), COLOUR_EX(c1),
#define STRIPES_V3(n,c0,c1,c2)				CMD_MAKE(Cmd_Stripesv,3), PAIR(n,COLOUR_EX(c0)), PAIR(COLOUR_EX(c1),COLOUR_EX(c2)),
#define STRIPES_V4(n,c0,c1,c2,c3)			CMD_MAKE(Cmd_Stripesv,4), PAIR(n,COLOUR_EX(c0)), PAIR(COLOUR_EX(c1),COLOUR_EX(c2)), COLOUR_EX(c3),
#define TRIANGLE(x0,y0,x1,y1)				CMD_MAKE(Cmd_Ext,CmdExt_Triangle), x0, y0, x1, y1,
#define VIEWBOX(x0,y0,x1,y1)				CMD_MAKE(Cmd_Ext,CmdExt_Viewbox), x0, y0, x1, y1,

local const signed char raw_flag_data[] =
{
	FLAG_BEGIN(None)
		ASPECT_RATIO(2,1)
	FLAG_END(None)
#include "MapFlag.h"
	Cmd_EOF
};
local const char* raw_flag[Flag_Count];

typedef struct _MAPFLAG_XFORM
{
	double xscalex;
	double xscaley;
	double yscalex;
	double yscaley;
	double xoffset;	// Includes the rounding 0.5
	double yoffset;	// Includes the rounding 0.5
	double aspect;
}
MAPFLAG_XFORM;

typedef struct _MAPFLAG
{
	HDC hdc;
	const char* p;
	HBRUSH* hbrush;
	HPEN* hpen;
	enum FlagColour brush;
	enum FlagColour brush_selected;
	MAPFLAG_XFORM xform;
}
MAPFLAG;

local POINT DrawOrdinateXY(MAPFLAG* flag, double x, double y)
{
	POINT pt;
	pt.x = (LONG)(x * flag->xform.xscalex + y * flag->xform.xscaley + flag->xform.xoffset);
	pt.y = (LONG)(x * flag->xform.yscalex + y * flag->xform.yscaley + flag->xform.yoffset);
	return pt;
}

local HBRUSH DrawGetBrush(MAPFLAG* flag, enum FlagColour colour)
{
#define FLAG_COLOUR_LIST_RGB(z,r,g,b) PALETTERGB(0x##r##r,0x##g##g,0x##b##b),
	static const COLORREF rgb[FlagColour_Count] =
	{
	    FLAG_COLOUR_LIST(FLAG_COLOUR_LIST_RGB)
	};

	// If necessary create the brush and pen
	ASSERT((colour >= 0) && (colour < dimof(rgb)));
	if (flag->hbrush[colour] == NULL)
	{
		// Create the brush and pen in tandem
		ASSERT(flag->hpen[colour] == NULL);
		flag->hbrush[colour] = OS(CreateSolidBrush)(rgb[colour]);
		ASSERT(flag->hbrush[colour] != NULL);
		flag->hpen[colour] = OS(CreatePen)(PS_SOLID, 0, rgb[colour]);
		ASSERT(flag->hpen[colour] != NULL);
	}
	return flag->hbrush[colour];
}

local void DrawSelectBrush(MAPFLAG* flag)
{
	// Set the current brush
	if (flag->hdc != NULL)
	{
		ASSERT((flag->brush >= 0) && (flag->brush < FlagColour_Count));
		if (flag->brush != flag->brush_selected)
		{
			OS(SelectObject)(flag->hdc, DrawGetBrush(flag, flag->brush));
			OS(SelectObject)(flag->hdc, flag->hpen[flag->brush]);
			flag->brush_selected = flag->brush;
		}
	}
	return;
}

local void DrawRectangleEx(MAPFLAG* flag, double x0, double y0, double x1, double y1, enum FlagColour colour)
{
	// Set the current brush
	if (flag->hdc != NULL)
	{
		POINT xy0 = DrawOrdinateXY(flag, x0, y0);
		POINT xy1 = DrawOrdinateXY(flag, x1, y1);
		RECT rect;
		rect.left = xy0.x;
		rect.bottom = xy0.y;
		rect.right = xy1.x;
		rect.top = xy1.y;
		VERIFY(OS(FillRect)(flag->hdc, &rect, DrawGetBrush(flag, colour)));
	}
	return;
}

//----------------------------------------------------------------------------

local void DrawCommand(MAPFLAG* flag);
local void DrawFlag(MAPFLAG* flag);

local void DrawColour(MAPFLAG* flag, BYTE arg)
{
    flag->brush = arg;
	return;
}

local void DrawBlockh(MAPFLAG* flag, BYTE arg)
{
    signed y0 = *(flag->p++);
    signed y1 = *(flag->p++);
    DrawRectangleEx(flag, -120, y0, 120, y1, arg);
	return;
}

local void DrawBlockv(MAPFLAG* flag, BYTE arg)
{
    signed x0 = *(flag->p++);
    signed x1 = *(flag->p++);
    DrawRectangleEx(flag, x0, -120, x1, 120, arg);
	return;
}

local LONG DrawRadius(double rx, double ry)
{
	return (LONG)(sqrt(rx * rx + ry * ry) + 0.5);
}

local void DrawCircle(MAPFLAG* flag, BYTE arg)
{
    double cx = *(flag->p++);
    double cy = *(flag->p++);
    double ry = arg ? arg : *(flag->p++);
    double rx = ry / flag->xform.aspect;
	ASSERT(rx > 0);
    ASSERT(ry > 0);
	if (flag->hdc)
	{
		POINT cxy = DrawOrdinateXY(flag, cx, cy);
		LONG dx = DrawRadius(rx * flag->xform.xscalex, ry * flag->xform.xscaley);
		LONG dy = DrawRadius(rx * flag->xform.yscalex, ry * flag->xform.yscaley);
		DrawSelectBrush(flag);
		VERIFY(OS(Ellipse)(flag->hdc, cxy.x - dx, cxy.y - dy, cxy.x + dx, cxy.y + dy));
	}
	return;
}

local void DrawCross(MAPFLAG* flag, BYTE arg)
{
    signed w = *(flag->p++);
    signed x = *(flag->p++);
    signed h = (signed)(w * flag->xform.aspect + 0.5);
    ASSERT(w > 0);
    ASSERT(h > 0);
    DrawRectangleEx(flag, x - w, -120, x + w, 120, arg);
    DrawRectangleEx(flag, -120, -h, 120, h, arg);
	return;
}

local void DrawCurve(MAPFLAG* flag, BYTE arg)
{
	unsigned bezier[64];
	POINT pt[dimof(bezier)][3];
	unsigned i, n = 1;
	enum CmdCurve curve = CmdCurve_Linear;
	unsigned triple;
	pt[0][0].x = *(flag->p++);
	pt[0][0].y = *(flag->p++);
	ASSERT(*flag->p != CmdCurve_End);
	do
	{
		switch (*flag->p)
		{
		case CmdCurve_Linear:
		case CmdCurve_Smooth:
		case CmdCurve_Cubic:
			ASSERT(curve != *flag->p);
			curve = *(flag->p++);
			break;
		default:
			ASSERT(n < dimof(bezier));
			switch (curve)
			{
			case CmdCurve_Linear:
				bezier[n] = 0;
				pt[n][0].x = *(flag->p++);
				pt[n][0].y = *(flag->p++);
				break;
			case CmdCurve_Smooth:
				ASSERT(n > 0);
				ASSERT(bezier[n-1]);
				bezier[n] = 1;
				pt[n][2].x = *(flag->p++);
				pt[n][2].y = *(flag->p++);
				pt[n][1].x = -pt[n-1][2].x;
				pt[n][1].y = -pt[n-1][2].y;
				pt[n][0].x = *(flag->p++);
				pt[n][0].y = *(flag->p++);
				break;
			case CmdCurve_Cubic:
				bezier[n] = 1;
				pt[n][1].x = *(flag->p++);
				triple = *(const BYTE*)(flag->p++);
				triple = (triple << 8) | *(const BYTE*)(flag->p++);
				pt[n][1].y = (triple >> 10) - 15;
				pt[n][2].x = ((triple >> 5) & 31) - 15;
				pt[n][2].y = (triple & 31) - 15;
				pt[n][0].x = *(flag->p++);
				pt[n][0].y = *(flag->p++);
				break;
			}
			n++;
			break;
		}
	}
	while (*flag->p != CmdCurve_End);
	flag->p++;
	if (flag->hdc != NULL)
	{
		POINT bz[3];
		VERIFY(OS(BeginPath)(flag->hdc));
		bz[0] = DrawOrdinateXY(flag, pt[0][0].x, pt[0][0].y);
		VERIFY(OS(MoveToEx)(flag->hdc, bz[0].x, bz[0].y, NULL));
		for (i = 1; i < n; ++i)
		{
			bz[2] = DrawOrdinateXY(flag, pt[i][0].x, pt[i][0].y);
			if (bezier[i])
			{
				bz[0] = DrawOrdinateXY(flag, pt[i-1][0].x + pt[i][1].x, pt[i-1][0].y + pt[i][1].y);
				bz[1] = DrawOrdinateXY(flag, pt[i][0].x + pt[i][2].x, pt[i][0].y + pt[i][2].y);
				VERIFY(OS(PolyBezierTo)(flag->hdc, bz, dimof(bz)));
			}
			else
			{
				VERIFY(OS(LineTo)(flag->hdc, bz[2].x, bz[2].y));
			}
		}
		if (arg)
		{
			// Mirrored too
			signed mx = pt[0][0].x * 2;
			for (i = n - 1; i > 0; --i)
			{
				bz[2] = DrawOrdinateXY(flag, mx - pt[i-1][0].x, pt[i-1][0].y);
				if (bezier[i])
				{
					bz[0] = DrawOrdinateXY(flag, mx - pt[i][0].x - pt[i][2].x, pt[i][0].y + pt[i][2].y);
					bz[1] = DrawOrdinateXY(flag, mx - pt[i-1][0].x - pt[i][1].x, pt[i-1][0].y + pt[i][1].y);
					VERIFY(OS(PolyBezierTo)(flag->hdc, bz, dimof(bz)));
				}
				else
				{
					VERIFY(OS(LineTo)(flag->hdc, bz[2].x, bz[2].y));
				}
			}
		}
		VERIFY(OS(EndPath)(flag->hdc));
		DrawSelectBrush(flag);
		VERIFY(OS(FillPath)(flag->hdc));
	}
	return;
}

local void DrawPolygon(MAPFLAG* flag, POINT pt[6], unsigned n, double sx, double sy)
{
	unsigned i;
	ASSERT(flag->hdc != NULL);
	ASSERT(n > 2);
	for (i = 0; i < n; i++)
	{
		pt[i] = DrawOrdinateXY(flag, pt[i].x * sx, pt[i].y * sy);
	}
	VERIFY(OS(Polygon)(flag->hdc, pt, n));
	return;
}

local void DrawDiagonalEx(MAPFLAG* flag, signed w, signed mirror)
{
	POINT pt[6];
    pt[0].x = -120;
	pt[0].y = -120;
	pt[1].x = -120 + w;
	pt[1].y = -120;
    pt[2].x = +120;
	pt[2].y = +120 - w;
    pt[3].x = +120;
	pt[3].y = +120;
    pt[4].x = +120 - w;
	pt[4].y = +120;
    pt[5].x = -120;
	pt[5].y = -120 + w;
	DrawPolygon(flag, pt, dimof(pt), mirror, 1);
	return;
}

local void DrawDiagonal(MAPFLAG* flag, BYTE arg)
{
    signed w = *(flag->p++);
	ASSERT(arg != 0);
	if (flag->hdc != NULL)
	{
		DrawSelectBrush(flag);
		if (arg & 1)
		{
			DrawDiagonalEx(flag, w, -1);
		}
		if (arg & 2)
		{
			DrawDiagonalEx(flag, w, +1);
		}
	}
	return;
}

local void DrawFlood(MAPFLAG* flag, BYTE arg)
{
    DrawRectangleEx(flag, -120, -120, +120, +120, arg);
	return;
}

local void DrawForr(MAPFLAG* flag, BYTE arg)
{
	unsigned i, n = arg ? arg : *(const BYTE*)(flag->p++);
	double a = M_PI * 2 / n;
	const char* stack = flag->p;
	MAPFLAG_XFORM xform = flag->xform;
	ASSERT(n > 1);
	flag->xform.aspect = 1;
	for (i = 0; i < n; ++i)
	{
		double cos_a = cos(a * i);
		double sin_a = sin(a * i);
		flag->xform.xscalex = xform.xscalex * cos_a / xform.aspect;
		flag->xform.xscaley = xform.xscalex * -sin_a / xform.aspect;
		flag->xform.yscalex = xform.yscaley * sin_a;
		flag->xform.yscaley = xform.yscaley * cos_a;
		flag->p = stack;
		DrawCommand(flag);
	}
	flag->xform = xform;
	return;
}

local void DrawForsEx(MAPFLAG* flag, double sx, double sy)
{
	MAPFLAG_XFORM xform = flag->xform;
	flag->xform.xscalex *= sx;
	flag->xform.yscaley *= sy;
	DrawCommand(flag);
	flag->xform = xform;
	return;
}

local void DrawFors(MAPFLAG* flag, BYTE arg)
{
	double sx = arg - 1;
	double sy = *(flag->p++) / 60.0;
	if (flag->hdc)
	{
		if ((sx < 0) || (sy < 0))
		{
			const char* stack = flag->p;
			DrawForsEx(flag, fabs(sx), fabs(sy));
			flag->p = stack;
		}
		DrawForsEx(flag, sx, sy);
	}
	return;
}

local void DrawFort(MAPFLAG* flag, BYTE arg)
{
	double dx = *(flag->p++) * flag->xform.xscalex;
	double dy = *(flag->p++) * flag->xform.yscaley;
	if (flag->hdc)
	{
		MAPFLAG_XFORM xform = flag->xform;
		const char* stack = flag->p;
		unsigned i;
		ASSERT(arg > 1);
		DrawCommand(flag);
		for (i = 1; i < arg; ++i)
		{
			flag->xform.xoffset += dx;
			flag->xform.yoffset += dy;
			flag->p = stack;
			DrawCommand(flag);
		}
		flag->xform = xform;
	}
	return;
}

local void DrawRectangle(MAPFLAG* flag)
{
    signed x0 = *(flag->p++);
    signed y0 = *(flag->p++);
    signed x1 = *(flag->p++);
    signed y1 = *(flag->p++);
	if (flag->hdc != NULL)
	{
		POINT pt[4];
		pt[0].x = x0;
		pt[0].y = y0;
		pt[1].x = x0;
		pt[1].y = y1;
		pt[2].x = x1;
		pt[2].y = y1;
		pt[3].x = x1;
		pt[3].y = y0;
		DrawSelectBrush(flag);
		DrawPolygon(flag, pt, dimof(pt), 1, 1);
	}
	return;
}

local void DrawStarEx(MAPFLAG* flag, unsigned n, double r1, double r2, double rot)
{
    double x = *(flag->p++);
    double y = *(flag->p++);
    if (r2 == 0)
    {
        r2 = r1 * 0.385;
    }
    ASSERT(n >= 3);
    ASSERT(r1 > 0);
    ASSERT(r2 > 0);
	if (flag->hdc != NULL)
	{
		POINT pt[40];
		unsigned i;
		ASSERT((n * 2) <= dimof(pt));
		for (i = 0; i < (n * 2); i++)
		{
			double angle = M_PI * i / n + rot;
			double ry = r1;
			double rx = ry / flag->xform.aspect;
			r1 = r2;
			r2 = ry;
			pt[i] = DrawOrdinateXY(flag, x + sin(angle) * rx, y + cos(angle) * ry);
		}
		DrawSelectBrush(flag);
		VERIFY(OS(Polygon)(flag->hdc, pt, i));
	}
	return;
}

local void DrawStar(MAPFLAG* flag, BYTE arg)
{
    unsigned n = arg + 4;
    double r1 = *(flag->p++);
    double r2 = *(flag->p++);
    double rot = *(flag->p++) * M_PI / 120.0;
	DrawStarEx(flag, n, r1, r2, rot);
	return;
}

local void DrawStar5(MAPFLAG* flag, BYTE arg)
{
    double r;
	if (arg)
	{
		// Small star
		r = arg + 1;
	}
	else
	{
		// Large star
		r = *(flag->p++);
		ASSERT(r > 16);
	}
	DrawStarEx(flag, 5, r, 0, 0);
	return;
}

local void DrawStripesEx(MAPFLAG* flag, int horz, unsigned colours)
{
    unsigned stripes;
    unsigned i;
    double x, y, z, dx, dy, dz, sx, sy;
	BYTE arg[6];
	BYTE got = 0;
    ASSERT((colours >= 2) && (colours <= 4));
	do
	{
		ASSERT(got < sizeof(arg));
		arg[got++] = *(const BYTE*)flag->p & 0x0F;
		arg[got++] = *(const BYTE*)flag->p >> 4;
		flag->p++;
	}
	while (got <= colours);
	ASSERT(got <= dimof(arg));
    stripes = arg[0];
    ASSERT(stripes >= 2);
	x = -120.0;
    z = 241.0 / stripes;
	dz = (240.0 - z) / (stripes - 1);
	ASSERT(dz < z);
    if (horz)
    {
        y = 120.0 - z;
        dx = 0.0;
        dy = -dz;
        sx = 240.0;
        sy = z;
    }
    else
    {
        y = -120.0;
        dx = dz;
        dy = 0.0;
        sx = z;
        sy = 240.0;
    }
    for (i = 0; i < stripes; i++)
    {
		unsigned colour = (i % colours) + 1;
		DrawRectangleEx(flag, x, y, x + sx, y + sy, arg[colour]);
        x += dx;
        y += dy;
    }
	return;
}

local void DrawStripesh(MAPFLAG* flag, BYTE arg)
{
	DrawStripesEx(flag, 1, arg);
	return;
}

local void DrawStripesv(MAPFLAG* flag, BYTE arg)
{
	DrawStripesEx(flag, 0, arg);
	return;
}

local void DrawJack(MAPFLAG* flag)
{
	if (flag->hdc)
	{
		MAPFLAG subflag = *flag;
		subflag.xform.xscalex *= 0.5;
		subflag.xform.yscaley *= 0.5;
		subflag.xform.xoffset *= 0.5;
		subflag.xform.yoffset *= 0.5;
		subflag.p = raw_flag[Flag_GB];
		DrawFlag(&subflag);
	}
	return;
}

local void DrawTriangle(MAPFLAG* flag)
{
	POINT pt[3];
    pt[0].x = *(flag->p++);
    pt[1].y = *(flag->p++);
    pt[1].x = *(flag->p++);
    pt[0].y = *(flag->p++);
	if (flag->hdc != NULL)
	{
		unsigned i = (pt[1].x < pt[0].x) ^ (pt[1].y < pt[0].y);
	    pt[2].x = pt[1 - i].x;
		pt[2].y = pt[i].y;
		DrawSelectBrush(flag);
		DrawPolygon(flag, pt, dimof(pt), 1, 1);
	}
	return;
}

local void DrawViewbox(MAPFLAG* flag)
{
    signed x0 = *(flag->p++);
    signed y0 = *(flag->p++);
    signed x1 = *(flag->p++);
    signed y1 = *(flag->p++);
	if (flag->hdc)
	{
		double sx = (x1 - x0) / 240.0;
		double sy = (y1 - y0) / 240.0;
		double tx = (x0 + x1) * 0.5;
		double ty = (y0 + y1) * 0.5;
		flag->xform.xoffset += flag->xform.xscalex * tx;
		flag->xform.yoffset += flag->xform.yscaley * ty;
		flag->xform.xscalex *= sx;
		flag->xform.yscaley *= sy;
		flag->xform.aspect *= sx / sy;
	}
	return;
}

local void DrawExt(MAPFLAG* flag, BYTE arg)
{
	typedef void (*JUMP)(MAPFLAG* flag);
	static const JUMP jump[] =
	{
		NULL // CmdExt_End
#define CMDEXT_LIST_JUMP(x) , Draw##x
        CMDEXT_LIST(CMDEXT_LIST_JUMP)
	};
	ASSERT(arg < (sizeof(jump) / sizeof(*jump)));
	ASSERT(jump[arg]);
	jump[arg](flag);
	return;
}

local void DrawCommandEx(MAPFLAG* flag, enum Cmd cmd)
{
	typedef void (*JUMP)(MAPFLAG* flag, BYTE arg);
	static const JUMP jump[] =
	{
		DrawExt // Cmd_Ext
#define CMD_LIST_JUMP(x) , Draw##x
        CMD_LIST(CMD_LIST_JUMP)
	};
	jump[CMD_GET_CMD(cmd)](flag, CMD_GET_ARG(cmd));
	return;
}

local void DrawCommand(MAPFLAG* flag)
{
	DrawCommandEx(flag, *(const BYTE*)(flag->p++));
	return;
}

local double DrawFlagAspectRatio(const char* p)
{
	// Work out a flag's aspect ratio based on the FLAG_ASPECT data at the beginning of the record
	return *(const BYTE*)p / 255.0 * (ASPECT_RATIO_MAX - ASPECT_RATIO_MIN) + ASPECT_RATIO_MIN;
}

double FlagAspectRatio(int flag)
{
	// Work out a flag's aspect ratio based on the FLAG_ASPECT data at the beginning of the record
	ASSERT((flag >= 0) && (flag < dimof(raw_flag)));
	return DrawFlagAspectRatio(raw_flag[flag]);
}

local void DrawFlag(MAPFLAG* flag)
{
	// Draw a flag into a new device context
	enum Cmd cmd;
	flag->xform.aspect = DrawFlagAspectRatio(flag->p++);
	flag->brush = FlagColour_W;
	flag->brush_selected = -1;
    cmd = *(const BYTE*)(flag->p++);
    while (cmd != Cmd_End)
    {
		DrawCommandEx(flag, cmd);
        cmd = *(const BYTE*)(flag->p++);
    }
	OS(SelectObject)(flag->hdc, OS(GetStockObject)(BLACK_PEN));
	OS(SelectObject)(flag->hdc, OS(GetStockObject)(WHITE_BRUSH));
	return;
}

void FlagDraw(HDC hdc, int width, int height, double aspect, int flag)
{
	HBRUSH hbrush[FlagColour_Count] = { 0 };
	HPEN hpen[FlagColour_Count] = { 0 };
	MAPFLAG fs = { 0 };
	int i;

	// Fill in a structure and forward on
	ASSERT((flag >= 0) && (flag < dimof(raw_flag)));
	fs.hdc = hdc;
	fs.p = raw_flag[flag];
	fs.hbrush = hbrush;
	fs.hpen = hpen;
	fs.xform.xscalex = width / 240.0;
	fs.xform.yscaley = height / -240.0;
	fs.xform.xoffset = width / 2.0 + 0.5;
	fs.xform.yoffset = height / 2.0 + 0.5;
	fs.xform.aspect = aspect;
	DrawFlag(&fs);
	for (i = 0; i < dimof(hbrush); i++)
	{
		// Clean up used brushes/pens
		if (hbrush[i] != NULL)
			VERIFY(OS(DeleteObject)(hbrush[i]));
		if (hpen[i] != NULL)
			VERIFY(OS(DeleteObject)(hpen[i]));
	}
	return;
}

void FlagDecompress(void)
{
	MAPFLAG fs = { 0 };
	int i;

	// Fill in a structure and forward on
	fs.p = raw_flag_data;
	for (i = 0; i < Flag_Count; i++)
	{
		raw_flag[i] = fs.p;
		DrawFlag(&fs);
	}
	ASSERT(*fs.p == Cmd_EOF);
	return;
}
