//----------------------------------------------------------------------------
//	SOURCE:	Map.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Tooltips for Map2 demo application
//----------------------------------------------------------------------------

#include "Map.h"

// The current tooltip
local TOOLINFO toolinfo;
local int tooltip_region;

HWND TooltipCreate(HWND hparent)
{
	// Install our tooltip
	static const TCHAR caption[] = _T("Map");
	HWND hwnd = OS(CreateWindowEx)(0, TOOLTIPS_CLASS, caption, TTS_NOPREFIX,
								   CW_USEDEFAULT,CW_USEDEFAULT, CW_USEDEFAULT,CW_USEDEFAULT,
								   hparent, NULL, NULL, NULL);
	if (hwnd != NULL)
	{
		toolinfo.cbSize = sizeof(toolinfo);
		toolinfo.uFlags = TTF_SUBCLASS|TTF_TRACK|TTF_ABSOLUTE|TTF_TRANSPARENT|TTF_IDISHWND;
		toolinfo.hwnd = hwnd;
		toolinfo.uId = (UINT)hparent;
		OS(GetClientRect)(hwnd, &toolinfo.rect);
		toolinfo.hinst = NULL;
		toolinfo.lpszText = (LPTSTR)caption;
		OS(SendMessage)(hwnd, TTM_ADDTOOL, 0, (LPARAM)&toolinfo);
		OS(SendMessage)(hwnd, TTM_TRACKACTIVATE, TRUE, (LPARAM)&toolinfo);
		OS(SendMessage)(hwnd, TTM_SETMAXTIPWIDTH, 0, 0x7FFF);
	}
	tooltip_region = -1;
	return hwnd;
}

void TooltipTrack(LPPOINT pt, int region)
{
	// Track the position of the mouse pointer
	if (region != tooltip_region)
	{
		// The region has changed
		if (tooltip_region < 0)
		{
			// Ask Windows to tell us when we leave the client area
			TRACKMOUSEEVENT tme;
			tme.cbSize = sizeof(tme);
			tme.dwFlags = TME_LEAVE;
			tme.hwndTrack = (HWND)toolinfo.uId;
			VERIFY(OS(TrackMouseEvent)(&tme));
		}
		if (region <= 0)
		{
			// No tooltip
			OS(ShowWindow)(toolinfo.hwnd, SW_HIDE);
		}
		else
		{
			// The text has changed
			toolinfo.lpszText = ReportTooltip(region);
			OS(SendMessage)(toolinfo.hwnd, TTM_UPDATETIPTEXT, 0, (LPARAM)&toolinfo);
		}
		tooltip_region = region;
	}
	if (tooltip_region > 0)
	{
		// Track the position
		ASSERT(pt != NULL);
		OS(SendMessage)(toolinfo.hwnd, TTM_TRACKPOSITION, 0, MAKELPARAM(pt->x - 16, pt->y + 24));
	}
	return;
}
